@@ -1084,7 +1084,9 @@ clutter_animator_get_timeline (ClutterAnimator *animator)
  * Start the ClutterAnimator, this is a thin wrapper that rewinds
  * and starts the animators current timeline.
  *
- * Return value: the #ClutterTimeline that drives the animator.
+ * Return value: (transfer none): the #ClutterTimeline that drives
+ *   the animator. The returned timeline is owned by the #ClutterAnimator
+ *   and it should not be unreferenced
  *
  * Since: 1.2
  */
@@ -1352,7 +1354,7 @@ clutter_animator_set_key (ClutterAnimator *animator,
  * Returns a list of pointers to opaque structures with accessor functions
  * that describe the keys added to an animator.
  *
- * Return value: (transfer container) (element-type ClutterAnimatorKey): a
+ * Return value: (transfer container) (element-type Clutter.AnimatorKey): a
  *   list of #ClutterAnimatorKey<!-- -->s; the contents of the list are owned
  *   by the #ClutterAnimator, but you should free the returned list when done,
  *   using g_list_free()
@@ -1443,23 +1445,14 @@ again:
         }
     }
 
-  if (object)
-    {
-      GHashTableIter iter;
-      gpointer key, value;
+  /* clear off cached state for all properties, this is regenerated in a
+   * correct state by animation_animator_started
+   */
+  g_hash_table_remove_all (priv->properties);
 
-again2:
-      g_hash_table_iter_init (&iter, priv->properties);
-      while (g_hash_table_iter_next (&iter, &key, &value))
-        {
-          PropObjectKey *prop_actor_key = key;
-          if (prop_actor_key->object == object)
-            {
-              g_hash_table_remove (priv->properties, key);
-              goto again2;
-            }
-        }
-    }
+  /* if the animator is already running reinitialize internal iterators */
+  if (clutter_timeline_is_playing (priv->timeline))
+    animation_animator_started (priv->timeline, animator);
 }
 
 /**
