@@ -48,6 +48,22 @@
  *   in both columns and rows.</para></listitem>
  * </itemizedlist>
  *
+ * <figure id="flow-layout-horizontal">
+ *   <title>Horizontal flow layout</title>
+ *   <para>The image shows a #ClutterFlowLayout with the
+ *   #ClutterFlowLayout:orientation propert set to
+ *   %CLUTTER_FLOW_HORIZONTAL.</para>
+ *   <graphic fileref="flow-layout-horizontal.png" format="PNG"/>
+ * </figure>
+ *
+ * <figure id="flow-layout-vertical">
+ *   <title>Vertical flow layout</title>
+ *   <para>The image shows a #ClutterFlowLayout with the
+ *   #ClutterFlowLayout:orientation propert set to
+ *   %CLUTTER_FLOW_VERTICAL.</para>
+ *   <graphic fileref="flow-layout-vertical.png" format="PNG"/>
+ * </figure>
+ *
  * #ClutterFlowLayout is available since Clutter 1.2
  */
 
@@ -88,6 +104,8 @@ struct _ClutterFlowLayoutPrivate
   /* per-line size */
   GArray *line_min;
   GArray *line_natural;
+  gfloat req_width;
+  gfloat req_height;
 
   guint line_count;
 
@@ -201,21 +219,23 @@ clutter_flow_layout_get_preferred_width (ClutterLayoutManager *manager,
   item_y = 0;
 
   /* clear the line width arrays */
-  if (priv->orientation == CLUTTER_FLOW_VERTICAL && for_height > 0)
-    {
-      if (priv->line_min != NULL)
-        g_array_free (priv->line_min, TRUE);
+  if (priv->line_min != NULL)
+    g_array_free (priv->line_min, TRUE);
 
-      if (priv->line_natural != NULL)
-        g_array_free (priv->line_natural, TRUE);
+  if (priv->line_natural != NULL)
+    g_array_free (priv->line_natural, TRUE);
 
-      priv->line_min = g_array_sized_new (FALSE, FALSE,
+  priv->line_min = g_array_sized_new (FALSE, FALSE,
+                                      sizeof (gfloat),
+                                      16);
+  priv->line_natural = g_array_sized_new (FALSE, FALSE,
                                           sizeof (gfloat),
                                           16);
-      priv->line_natural = g_array_sized_new (FALSE, FALSE,
-                                              sizeof (gfloat),
-                                              16);
-    }
+
+  if (children)
+    line_count = 1;
+
+  max_min_width = max_natural_width = 0;
 
   for (l = children; l != NULL; l = l->next)
     {
@@ -271,6 +291,8 @@ clutter_flow_layout_get_preferred_width (ClutterLayoutManager *manager,
           max_min_width = MAX (max_min_width, child_min);
           max_natural_width = MAX (max_natural_width, child_natural);
 
+          total_min_width += max_min_width;
+          total_natural_width += max_natural_width;
           line_count += 1;
         }
     }
@@ -300,6 +322,7 @@ clutter_flow_layout_get_preferred_width (ClutterLayoutManager *manager,
         }
 
       priv->line_count = line_count;
+
       if (priv->line_count > 0)
         {
           gfloat total_spacing;
@@ -310,6 +333,32 @@ clutter_flow_layout_get_preferred_width (ClutterLayoutManager *manager,
           total_natural_width += total_spacing;
         }
     }
+  else
+    {
+      g_array_append_val (priv->line_min, line_min_width);
+      g_array_append_val (priv->line_natural, line_natural_width);
+
+      priv->line_count = line_count;
+
+      if (priv->line_count > 0)
+        {
+          gfloat total_spacing;
+
+          total_spacing = priv->col_spacing * (priv->line_count - 1);
+
+          total_min_width += total_spacing;
+          total_natural_width += total_spacing;
+        }
+    }
+
+  CLUTTER_NOTE (LAYOUT,
+                "Flow[w]: %d lines (%d per line): w [ %.2f, %.2f ] for h %.2f",
+                n_rows, priv->line_count,
+                total_min_width,
+                total_natural_width,
+                for_height);
+
+  priv->req_height = for_height;
 
   if (min_width_p)
     *min_width_p = total_min_width;
@@ -347,21 +396,23 @@ clutter_flow_layout_get_preferred_height (ClutterLayoutManager *manager,
   item_x = 0;
 
   /* clear the line height arrays */
-  if (priv->orientation == CLUTTER_FLOW_HORIZONTAL && for_width > 0)
-    {
-      if (priv->line_min != NULL)
-        g_array_free (priv->line_min, TRUE);
+  if (priv->line_min != NULL)
+    g_array_free (priv->line_min, TRUE);
 
-      if (priv->line_natural != NULL)
-        g_array_free (priv->line_natural, TRUE);
+  if (priv->line_natural != NULL)
+    g_array_free (priv->line_natural, TRUE);
 
-      priv->line_min = g_array_sized_new (FALSE, FALSE,
+  priv->line_min = g_array_sized_new (FALSE, FALSE,
+                                      sizeof (gfloat),
+                                      16);
+  priv->line_natural = g_array_sized_new (FALSE, FALSE,
                                           sizeof (gfloat),
                                           16);
-      priv->line_natural = g_array_sized_new (FALSE, FALSE,
-                                              sizeof (gfloat),
-                                              16);
-    }
+
+  if (children)
+    line_count = 1;
+
+  max_min_height = max_natural_height = 0;
 
   for (l = children; l != NULL; l = l->next)
     {
@@ -417,6 +468,9 @@ clutter_flow_layout_get_preferred_height (ClutterLayoutManager *manager,
           max_min_height = MAX (max_min_height, child_min);
           max_natural_height = MAX (max_natural_height, child_natural);
 
+          total_min_height += max_min_height;
+          total_natural_height += max_natural_height;
+
           line_count += 1;
         }
     }
@@ -456,6 +510,32 @@ clutter_flow_layout_get_preferred_height (ClutterLayoutManager *manager,
           total_natural_height += total_spacing;
         }
     }
+  else
+    {
+      g_array_append_val (priv->line_min, line_min_height);
+      g_array_append_val (priv->line_natural, line_natural_height);
+
+      priv->line_count = line_count;
+
+      if (priv->line_count > 0)
+        {
+          gfloat total_spacing;
+
+          total_spacing = priv->col_spacing * priv->line_count;
+
+          total_min_height += total_spacing;
+          total_natural_height += total_spacing;
+        }
+    }
+
+  CLUTTER_NOTE (LAYOUT,
+                "Flow[h]: %d lines (%d per line): w [ %.2f, %.2f ] for h %.2f",
+                n_columns, priv->line_count,
+                total_min_height,
+                total_natural_height,
+                for_width);
+
+  priv->req_width = for_width;
 
   if (min_height_p)
     *min_height_p = total_min_height;
@@ -483,6 +563,21 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
 
   clutter_actor_box_get_size (allocation, &avail_width, &avail_height);
 
+  /* blow the cached preferred size and re-compute with the given
+   * available size in case the FlowLayout wasn't given the exact
+   * size it requested
+   */
+  if ((priv->req_width > 0 && avail_width != priv->req_width) ||
+      (priv->req_height > 0 && avail_height != priv->req_height))
+    {
+      clutter_flow_layout_get_preferred_width (manager, container,
+                                               avail_height,
+                                               NULL, NULL);
+      clutter_flow_layout_get_preferred_height (manager, container,
+                                                avail_width,
+                                                NULL, NULL);
+    }
+
   items_per_line = compute_lines (CLUTTER_FLOW_LAYOUT (manager),
                                   avail_width, avail_height);
 
@@ -501,6 +596,8 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
       if (!CLUTTER_ACTOR_IS_VISIBLE (child))
         continue;
 
+      new_x = new_y = 0;
+
       if (priv->orientation == CLUTTER_FLOW_HORIZONTAL)
         {
           if (line_item_count == items_per_line && line_item_count > 0)
@@ -509,6 +606,9 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
                                        gfloat,
                                        line_index);
 
+              if (line_index >= 0)
+                item_y += priv->row_spacing;
+
               line_item_count = 0;
               line_index += 1;
 
@@ -529,7 +629,7 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
               clutter_actor_get_preferred_width (child, item_height,
                                                  &child_min,
                                                  &child_natural);
-              item_width = MIN (item_width, child_min);
+              item_width = MIN (item_width, child_natural);
 
               clutter_actor_get_preferred_height (child, item_width,
                                                   &child_min,
@@ -545,6 +645,9 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
                                        gfloat,
                                        line_index);
 
+              if (line_index >= 0)
+                item_x += priv->col_spacing;
+
               line_item_count = 0;
               line_index += 1;
 
@@ -565,7 +668,7 @@ clutter_flow_layout_allocate (ClutterLayoutManager   *manager,
               clutter_actor_get_preferred_width (child, item_height,
                                                  &child_min,
                                                  &child_natural);
-              item_width = MIN (item_width, child_min);
+              item_width = MIN (item_width, child_natural);
 
               clutter_actor_get_preferred_height (child, item_width,
                                                   &child_min,
@@ -602,6 +705,7 @@ clutter_flow_layout_set_container (ClutterLayoutManager *manager,
                                    ClutterContainer     *container)
 {
   ClutterFlowLayoutPrivate *priv = CLUTTER_FLOW_LAYOUT (manager)->priv;
+  ClutterLayoutManagerClass *parent_class;
 
   priv->container = container;
 
@@ -618,6 +722,9 @@ clutter_flow_layout_set_container (ClutterLayoutManager *manager,
       clutter_actor_set_request_mode (CLUTTER_ACTOR (priv->container),
                                       request_mode);
     }
+
+  parent_class = CLUTTER_LAYOUT_MANAGER_CLASS (clutter_flow_layout_parent_class);
+  parent_class->set_container (manager, container);
 }
 
 static void
@@ -772,8 +879,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_enum ("orientation",
-                             "Orientation",
-                             "The orientation of the layout",
+                             P_("Orientation"),
+                             P_("The orientation of the layout"),
                              CLUTTER_TYPE_FLOW_ORIENTATION,
                              CLUTTER_FLOW_HORIZONTAL,
                              CLUTTER_PARAM_READWRITE |
@@ -789,9 +896,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_boolean ("homogeneous",
-                                "Homogeneous",
-                                "Whether each item should receive the "
-                                "same allocation",
+                                P_("Homogeneous"),
+                                P_("Whether each item should receive the same allocation"),
                                 FALSE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_HOMOGENEOUS, pspec);
@@ -806,8 +912,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("column-spacing",
-                              "Column Spacing",
-                              "The spacing between columns",
+                              P_("Column Spacing"),
+                              P_("The spacing between columns"),
                               0.0, G_MAXFLOAT,
                               0.0,
                               CLUTTER_PARAM_READWRITE);
@@ -825,8 +931,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("row-spacing",
-                              "Row Spacing",
-                              "The spacing between rows",
+                              P_("Row Spacing"),
+                              P_("The spacing between rows"),
                               0.0, G_MAXFLOAT,
                               0.0,
                               CLUTTER_PARAM_READWRITE);
@@ -842,8 +948,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("min-column-width",
-                              "Minimum Column Width",
-                              "Minimum width for each column",
+                              P_("Minimum Column Width"),
+                              P_("Minimum width for each column"),
                               0.0, G_MAXFLOAT,
                               0.0,
                               CLUTTER_PARAM_READWRITE);
@@ -860,8 +966,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("max-column-width",
-                              "Maximum Column Width",
-                              "Maximum width for each column",
+                              P_("Maximum Column Width"),
+                              P_("Maximum width for each column"),
                               -1.0, G_MAXFLOAT,
                               -1.0,
                               CLUTTER_PARAM_READWRITE);
@@ -877,8 +983,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("min-row-height",
-                              "Minimum Row Height",
-                              "Minimum height for each row",
+                              P_("Minimum Row Height"),
+                              P_("Minimum height for each row"),
                               0.0, G_MAXFLOAT,
                               0.0,
                               CLUTTER_PARAM_READWRITE);
@@ -895,8 +1001,8 @@ clutter_flow_layout_class_init (ClutterFlowLayoutClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_float ("max-row-height",
-                              "Maximum Row Height",
-                              "Maximum height for each row",
+                              P_("Maximum Row Height"),
+                              P_("Maximum height for each row"),
                               -1.0, G_MAXFLOAT,
                               -1.0,
                               CLUTTER_PARAM_READWRITE);
