@@ -35,7 +35,6 @@
 #include <gmodule.h>
 
 #include "clutter-actor.h"
-#include "clutter-behaviour.h"
 #include "clutter-container.h"
 #include "clutter-debug.h"
 #include "clutter-enum-types.h"
@@ -150,7 +149,7 @@ clutter_script_get_type_from_class (const gchar *name)
  * clutter_script_enum_from_string:
  * @type: a #GType for an enumeration type
  * @string: the enumeration value as a string
- * @enum_value: return location for the enumeration value as an integer
+ * @enum_value: (out): return location for the enumeration value as an integer
  *
  * Converts an enumeration value inside @string into a numeric
  * value and places it into @enum_value.
@@ -487,8 +486,8 @@ clutter_script_parse_color (ClutterScript *script,
   return FALSE;
 }
 
-static const gchar *
-get_id_from_node (JsonNode *node)
+const gchar *
+_clutter_script_get_id_from_node (JsonNode *node)
 {
   JsonObject *object;
 
@@ -531,7 +530,7 @@ parse_children (ObjectInfo *oinfo,
       JsonNode *child = json_array_get_element (array, i);
       const gchar *id;
 
-      id = get_id_from_node (child);
+      id = _clutter_script_get_id_from_node (child);
       if (id)
         retval = g_list_prepend (retval, g_strdup (id));
     }
@@ -651,35 +650,6 @@ parse_signals (ClutterScript *script,
   return retval;
 }
 
-static GList *
-parse_behaviours (ObjectInfo *oinfo,
-                  JsonNode   *node)
-{
-  JsonArray *array;
-  GList *retval;
-  guint array_len, i;
-
-  if (JSON_NODE_TYPE (node) != JSON_NODE_ARRAY)
-    return NULL;
-
-  retval = oinfo->behaviours;
-
-  array = json_node_get_array (node);
-  array_len = json_array_get_length (array);
-
-  for (i = 0; i < array_len; i++)
-    {
-      JsonNode *child = json_array_get_element (array, i);
-      const gchar *id;
-
-      id = get_id_from_node (child);
-      if (id)
-        retval = g_list_prepend (retval, g_strdup (id));
-    }
-
-  return g_list_reverse (retval);
-}
-
 static ClutterTimeline *
 construct_timeline (ClutterScript *script,
                     JsonObject    *object)
@@ -983,14 +953,6 @@ clutter_script_parser_object_end (JsonParser *json_parser,
       json_object_remove_member (object, "children");
     }
 
-  if (json_object_has_member (object, "behaviours"))
-    {
-      val = json_object_get_member (object, "behaviours");
-      oinfo->behaviours = parse_behaviours (oinfo, val);
-
-      json_object_remove_member (object, "behaviours");
-    }
-
   if (json_object_has_member (object, "signals"))
     {
       val = json_object_get_member (object, "signals");
@@ -1115,7 +1077,7 @@ clutter_script_parse_node (ClutterScript *script,
                * definitions are parsed leaf-first we are guaranteed
                * to have a defined object at this point
                */
-              id = get_id_from_node (node);
+              id = _clutter_script_get_id_from_node (node);
               if (id == NULL || *id == '\0')
                 return FALSE;
 
@@ -1740,46 +1702,6 @@ apply_child_properties (ClutterScript    *script,
   oinfo->properties = unresolved;
 }
 
-static void
-apply_behaviours (ClutterScript *script,
-                  ObjectInfo    *oinfo)
-{
-  ClutterActor *actor = CLUTTER_ACTOR (oinfo->object);
-  GList *l, *unresolved;
-
-  unresolved = NULL;
-  for (l = oinfo->behaviours; l != NULL; l = l->next)
-    {
-      const gchar *name = l->data;
-      ObjectInfo *behaviour_info;
-      GObject *object = NULL;
-
-      behaviour_info = _clutter_script_get_object_info (script, name);
-      if (behaviour_info != NULL)
-        {
-          _clutter_script_construct_object (script, behaviour_info);
-          object = behaviour_info->object;
-        }
-
-      if (object == NULL)
-        {
-          unresolved = g_list_prepend (unresolved, g_strdup (name));
-          continue;
-        }
-
-      CLUTTER_NOTE (SCRIPT, "Applying behaviour '%s' to actor of type '%s'",
-                    name,
-                    g_type_name (G_OBJECT_TYPE (actor)));
-
-      clutter_behaviour_apply (CLUTTER_BEHAVIOUR (object), actor);
-    }
-
-  g_list_foreach (oinfo->behaviours, (GFunc) g_free, NULL);
-  g_list_free (oinfo->behaviours);
-
-  oinfo->behaviours = unresolved;
-}
-
 static void
 add_children (ClutterScript *script,
               ObjectInfo    *oinfo)
@@ -1839,9 +1761,6 @@ _clutter_script_check_unresolved (ClutterScript *script,
   if (oinfo->children != NULL && CLUTTER_IS_CONTAINER (oinfo->object))
     add_children (script, oinfo);
 
-  if (oinfo->behaviours != NULL && CLUTTER_IS_ACTOR (oinfo->object))
-    apply_behaviours (script, oinfo);
-
   /* this is a bit *eugh*, but it allows us to effectively make sure
    * that child and layout properties are parsed and applied to the
    * right child
@@ -1854,7 +1773,6 @@ _clutter_script_check_unresolved (ClutterScript *script,
       if (parent != NULL && CLUTTER_IS_CONTAINER (parent))
         {
           ClutterContainer *container = CLUTTER_CONTAINER (parent);
-          ClutterActor *actor = CLUTTER_ACTOR (oinfo->object);
           GList *children, *l;
 
           children = clutter_container_get_children (container);
@@ -1874,10 +1792,10 @@ _clutter_script_check_unresolved (ClutterScript *script,
                 continue;
 
               apply_child_properties (script, container,
-                                      actor,
+                                      CLUTTER_ACTOR (child),
                                       child_info);
               apply_layout_properties (script, container,
-                                       actor,
+                                       CLUTTER_ACTOR (child),
                                        child_info);
             }
 
@@ -1885,7 +1803,7 @@ _clutter_script_check_unresolved (ClutterScript *script,
         }
     }
 
-  if (oinfo->properties || oinfo->children || oinfo->behaviours)
+  if (oinfo->properties || oinfo->children)
     oinfo->has_unresolved = TRUE;
   else
     oinfo->has_unresolved = FALSE;
@@ -2032,6 +1950,15 @@ _clutter_script_construct_object (ClutterScript *script,
                                      params->len,
                                      (GParameter *) params->data);
 
+      /* by sinking the floating reference, we make sure that the reference
+       * count is correct whether the object is referenced from somewhere
+       * else too or only by this ClutterScript object.
+       */
+      if (!oinfo->is_toplevel)
+        {
+          g_object_ref_sink (oinfo->object);
+        }
+
       for (i = 0; i < params->len; i++)
         {
           GParameter *param = &g_array_index (params, GParameter, i);
