@@ -320,15 +320,13 @@ object_info_free (gpointer data)
         {
           if (oinfo->is_unmerged)
             {
-              if (oinfo->is_toplevel)
-                g_object_unref (oinfo->object);
-              else
-                {
-                  /* destroy every actor, unless it's the default stage */
-                  if (oinfo->is_stage_default == FALSE &&
-                      CLUTTER_IS_ACTOR (oinfo->object))
-                    clutter_actor_destroy (CLUTTER_ACTOR (oinfo->object));
-                }
+              /* destroy every actor, unless it's the default stage */
+              if (!oinfo->is_toplevel &&
+	          oinfo->is_stage_default == FALSE &&
+                  CLUTTER_IS_ACTOR (oinfo->object))
+                clutter_actor_destroy (CLUTTER_ACTOR (oinfo->object));
+
+              g_object_unref(oinfo->object);
             }
           else
             {
