@@ -53,8 +53,9 @@
 #include "config.h"
 #endif
 
-#include "clutter-backend.h"
 #include "clutter-stage.h"
+
+#include "clutter-backend.h"
 #include "clutter-main.h"
 #include "clutter-color.h"
 #include "clutter-util.h"
@@ -78,6 +79,25 @@ G_DEFINE_TYPE (ClutterStage, clutter_stage, CLUTTER_TYPE_GROUP);
 #define CLUTTER_STAGE_GET_PRIVATE(obj) \
 (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CLUTTER_TYPE_STAGE, ClutterStagePrivate))
 
+/* <private>
+ * ClutterStageHint:
+ * @CLUTTER_STAGE_NONE: No hint set
+ * @CLUTTER_STAGE_NO_CLEAR_ON_PAINT: When this hint is set, the stage
+ *   should not clear the viewport; this flag is useful when painting
+ *   fully opaque actors covering the whole visible area of the stage,
+ *   i.e. when no blending with the stage color happens over the whole
+ *   stage viewport
+ *
+ * A series of hints that enable or disable behaviours on the stage
+ */
+typedef enum { /*< prefix=CLUTTER_STAGE >*/
+  CLUTTER_STAGE_HINT_NONE = 0,
+
+  CLUTTER_STAGE_NO_CLEAR_ON_PAINT = 1 << 0
+} ClutterStageHint;
+
+#define STAGE_NO_CLEAR_ON_PAINT(s)      ((((ClutterStage *) (s))->priv->stage_hints & CLUTTER_STAGE_NO_CLEAR_ON_PAINT) != 0)
+
 struct _ClutterStagePrivate
 {
   /* the stage implementation */
@@ -85,6 +105,8 @@ struct _ClutterStagePrivate
 
   ClutterColor        color;
   ClutterPerspective  perspective;
+  CoglMatrix          projection;
+  int                 viewport[4];
   ClutterFog          fog;
 
   gchar              *title;
@@ -92,6 +114,10 @@ struct _ClutterStagePrivate
 
   GQueue             *event_queue;
 
+  ClutterStageHint    stage_hints;
+
+  gint                picks_per_frame;
+
   guint redraw_pending         : 1;
   guint is_fullscreen          : 1;
   guint is_cursor_visible      : 1;
@@ -100,6 +126,9 @@ struct _ClutterStagePrivate
   guint throttle_motion_events : 1;
   guint use_alpha              : 1;
   guint min_size_changed       : 1;
+  guint dirty_viewport         : 1;
+  guint dirty_projection       : 1;
+  guint have_valid_pick_buffer : 1;
 };
 
 enum
@@ -112,11 +141,12 @@ enum
   PROP_CURSOR_VISIBLE,
   PROP_PERSPECTIVE,
   PROP_TITLE,
-  PROP_USER_RESIZE,
+  PROP_USER_RESIZABLE,
   PROP_USE_FOG,
   PROP_FOG,
   PROP_USE_ALPHA,
-  PROP_KEY_FOCUS
+  PROP_KEY_FOCUS,
+  PROP_NO_CLEAR_HINT
 };
 
 enum
@@ -268,12 +298,22 @@ clutter_stage_allocate (ClutterActor           *self,
       klass = CLUTTER_ACTOR_CLASS (clutter_stage_parent_class);
       klass->allocate (self, &override, flags);
     }
+
+  clutter_actor_get_allocation_geometry (self, &geom);
+  _clutter_stage_set_viewport (CLUTTER_STAGE (self),
+                               0, 0, geom.width, geom.height);
+
+  /* Note: we don't assume that set_viewport will queue a full redraw
+   * since it may bail-out early if something preemptively set the
+   * viewport before the stage was really allocated its new size. */
+  clutter_actor_queue_redraw (self);
 }
 
 static void
 clutter_stage_paint (ClutterActor *self)
 {
   ClutterStagePrivate *priv = CLUTTER_STAGE (self)->priv;
+  CoglBufferBit clear_flags;
   CoglColor stage_color;
   guint8 real_alpha;
   CLUTTER_STATIC_TIMER (stage_clear_timer,
@@ -292,18 +332,22 @@ clutter_stage_paint (ClutterActor *self)
   /* we use the real alpha to clear the stage if :use-alpha is
    * set; the effect depends entirely on how the Clutter backend
    */
-  cogl_color_set_from_4ub (&stage_color,
-                           priv->color.red,
-                           priv->color.green,
-                           priv->color.blue,
-                           priv->use_alpha ? real_alpha
+  cogl_color_init_from_4ub (&stage_color,
+                            priv->color.red,
+                            priv->color.green,
+                            priv->color.blue,
+                            priv->use_alpha ? real_alpha
                                            : 255);
   cogl_color_premultiply (&stage_color);
 
+  clear_flags = COGL_BUFFER_BIT_DEPTH;
+  if (!STAGE_NO_CLEAR_ON_PAINT (self))
+    clear_flags |= COGL_BUFFER_BIT_COLOR;
+
   CLUTTER_TIMER_START (_clutter_uprof_context, stage_clear_timer);
-  cogl_clear (&stage_color,
-	      COGL_BUFFER_BIT_COLOR |
-	      COGL_BUFFER_BIT_DEPTH);
+
+  cogl_clear (&stage_color, clear_flags);
+
   CLUTTER_TIMER_STOP (_clutter_uprof_context, stage_clear_timer);
 
   if (priv->use_fog)
@@ -348,7 +392,8 @@ clutter_stage_realize (ClutterActor *self)
    * first paint (which will likely occur before the ConfigureNotify
    * is received)
    */
-  CLUTTER_SET_PRIVATE_FLAGS (self, CLUTTER_ACTOR_SYNC_MATRICES);
+  priv->dirty_viewport = TRUE;
+  priv->dirty_projection = TRUE;
 
   g_assert (priv->impl != NULL);
   is_realized = _clutter_stage_window_realize (priv->impl);
@@ -692,6 +737,16 @@ clutter_stage_real_queue_redraw (ClutterActor *actor,
   else
     CLUTTER_CONTEXT ()->redraw_count += 1;
 
+  /* We have an optimization in _clutter_do_pick to detect when the
+   * scene is static so we can cache a full, un-clipped pick buffer to
+   * avoid continuous pick renders.
+   *
+   * Currently the assumption is that actors queue a redraw when some
+   * state changes that affects painting *or* picking so we can use
+   * this point to invalidate any currently cached pick buffer.
+   */
+  _clutter_stage_set_pick_buffer_valid (stage, FALSE);
+
   /* If the backend can't do anything with redraw clips (e.g. it already knows
    * it needs to redraw everything anyway) then don't spend time transforming
    * any clip regions into stage coordinates... */
@@ -744,6 +799,9 @@ _clutter_stage_has_full_redraw_queued (ClutterStage *stage)
 {
   ClutterStageWindow *stage_window = _clutter_stage_get_window (stage);
 
+  if (CLUTTER_ACTOR_IN_DESTRUCTION (stage) || stage_window == NULL)
+    return FALSE;
+
   if (stage->priv->redraw_pending &&
       !_clutter_stage_window_has_redraw_clips (stage_window))
     return TRUE;
@@ -763,6 +821,72 @@ clutter_stage_real_delete_event (ClutterStage *stage,
   return TRUE;
 }
 
+static void
+clutter_stage_real_apply_transform (ClutterActor *stage,
+                                    CoglMatrix   *matrix)
+{
+  ClutterStagePrivate *priv = CLUTTER_STAGE (stage)->priv;
+  CoglMatrix perspective;
+  gfloat z_camera;
+  gfloat width, height;
+
+  /*
+   * In theory, we can compute the camera distance from screen as:
+   *
+   *   0.5 * tan (FOV)
+   *
+   * However, it's better to compute the z_camera from our projection
+   * matrix so that we get a 1:1 mapping at the screen distance. Consider
+   * the upper-left corner of the screen. It has object coordinates
+   * (0,0,0), so by the transform below, ends up with eye coordinate
+   *
+   *   x_eye = x_object / width - 0.5 = - 0.5
+   *   y_eye = (height - y_object) / width - 0.5 = 0.5
+   *   z_eye = z_object / width - z_camera = - z_camera
+   *
+   * From cogl_perspective(), we know that the projection matrix has
+   * the form:
+   *
+   *  (x, 0,  0, 0)
+   *  (0, y,  0, 0)
+   *  (0, 0,  c, d)
+   *  (0, 0, -1, 0)
+   *
+   * Applied to the above, we get clip coordinates of
+   *
+   *  x_clip = x * (- 0.5)
+   *  y_clip = y * 0.5
+   *  w_clip = - 1 * (- z_camera) = z_camera
+   *
+   * Dividing through by w to get normalized device coordinates, we
+   * have, x_nd = x * 0.5 / z_camera, y_nd = - y * 0.5 / z_camera.
+   * The upper left corner of the screen has normalized device coordinates,
+   * (-1, 1), so to have the correct 1:1 mapping, we have to have:
+   *
+   *   z_camera = 0.5 * x = 0.5 * y
+   *
+   * If x != y, then we have a non-uniform aspect ration, and a 1:1 mapping
+   * doesn't make sense.
+   */
+
+  cogl_matrix_init_identity (&perspective);
+  cogl_matrix_perspective (&perspective,
+                           priv->perspective.fovy,
+                           priv->perspective.aspect,
+                           priv->perspective.z_near,
+                           priv->perspective.z_far);
+
+  z_camera = 0.5f * perspective.xx;
+
+  clutter_actor_get_size (stage, &width, &height);
+
+  cogl_matrix_init_identity (matrix);
+  cogl_matrix_translate (matrix, -0.5f, -0.5f, -z_camera);
+  cogl_matrix_scale (matrix,
+                     1.0f / width, -1.0f / height, 1.0f / width);
+  cogl_matrix_translate (matrix, 0.0f, -1.0f * height, 0.0f);
+}
+
 static void
 clutter_stage_set_property (GObject      *object,
 			    guint         prop_id,
@@ -797,7 +921,7 @@ clutter_stage_set_property (GObject      *object,
       clutter_stage_set_title (stage, g_value_get_string (value));
       break;
 
-    case PROP_USER_RESIZE:
+    case PROP_USER_RESIZABLE:
       clutter_stage_set_user_resizable (stage, g_value_get_boolean (value));
       break;
 
@@ -817,6 +941,10 @@ clutter_stage_set_property (GObject      *object,
       clutter_stage_set_key_focus (stage, g_value_get_object (value));
       break;
 
+    case PROP_NO_CLEAR_HINT:
+      clutter_stage_set_no_clear_hint (stage, g_value_get_boolean (value));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -857,7 +985,7 @@ clutter_stage_get_property (GObject    *gobject,
       g_value_set_string (value, priv->title);
       break;
 
-    case PROP_USER_RESIZE:
+    case PROP_USER_RESIZABLE:
       g_value_set_boolean (value, priv->is_user_resizable);
       break;
 
@@ -877,6 +1005,15 @@ clutter_stage_get_property (GObject    *gobject,
       g_value_set_object (value, priv->key_focused_actor);
       break;
 
+    case PROP_NO_CLEAR_HINT:
+      {
+        gboolean hint =
+          (priv->stage_hints & CLUTTER_STAGE_NO_CLEAR_ON_PAINT) != 0;
+
+        g_value_set_boolean (value, hint);
+      }
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
       break;
@@ -959,6 +1096,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
   actor_class->show = clutter_stage_show;
   actor_class->hide = clutter_stage_hide;
   actor_class->queue_redraw = clutter_stage_real_queue_redraw;
+  actor_class->apply_transform = clutter_stage_real_apply_transform;
 
   /**
    * ClutterStage:fullscreen:
@@ -974,8 +1112,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 1.0
    */
   pspec = g_param_spec_boolean ("fullscreen-set",
-                                "Fullscreen Set",
-                                "Whether the main stage is fullscreen",
+                                P_("Fullscreen Set"),
+                                P_("Whether the main stage is fullscreen"),
                                 FALSE,
                                 CLUTTER_PARAM_READABLE);
   g_object_class_install_property (gobject_class,
@@ -991,9 +1129,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * and it might be deprecated at any later date.</para></warning>
    */
   pspec = g_param_spec_boolean ("offscreen",
-                                "Offscreen",
-                                "Whether the main stage should be "
-                                "rendered offscreen",
+                                P_("Offscreen"),
+                                P_("Whether the main stage should be rendered offscreen"),
                                 FALSE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class,
@@ -1005,9 +1142,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Whether the mouse pointer should be visible
    */
   pspec = g_param_spec_boolean ("cursor-visible",
-                                "Cursor Visible",
-                                "Whether the mouse pointer is visible "
-                                "on the main stage",
+                                P_("Cursor Visible"),
+                                P_("Whether the mouse pointer is visible on the main stage"),
                                 TRUE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class,
@@ -1021,13 +1157,12 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.4
    */
   pspec = g_param_spec_boolean ("user-resizable",
-                                "User Resizable",
-                                "Whether the stage is able to be resized "
-                                "via user interaction",
+                                P_("User Resizable"),
+                                P_("Whether the stage is able to be resized via user interaction"),
                                 FALSE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class,
-                                   PROP_USER_RESIZE,
+                                   PROP_USER_RESIZABLE,
                                    pspec);
   /**
    * ClutterStage:color:
@@ -1035,8 +1170,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * The color of the main stage.
    */
   pspec = clutter_param_spec_color ("color",
-                                    "Color",
-                                    "The color of the stage",
+                                    P_("Color"),
+                                    P_("The color of the stage"),
                                     &default_stage_color,
                                     CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_COLOR, pspec);
@@ -1050,8 +1185,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.8.2
    */
   pspec = g_param_spec_boxed ("perspective",
-                              "Perspective",
-                              "Perspective projection parameters",
+                              P_("Perspective"),
+                              P_("Perspective projection parameters"),
                               CLUTTER_TYPE_PERSPECTIVE,
                               CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class,
@@ -1066,8 +1201,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.4
    */
   pspec = g_param_spec_string ("title",
-                               "Title",
-                               "Stage Title",
+                               P_("Title"),
+                               P_("Stage Title"),
                                NULL,
                                CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_TITLE, pspec);
@@ -1082,8 +1217,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.6
    */
   pspec = g_param_spec_boolean ("use-fog",
-                                "Use Fog",
-                                "Whether to enable depth cueing",
+                                P_("Use Fog"),
+                                P_("Whether to enable depth cueing"),
                                 FALSE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_USE_FOG, pspec);
@@ -1097,8 +1232,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 1.0
    */
   pspec = g_param_spec_boxed ("fog",
-                              "Fog",
-                              "Settings for the depth cueing",
+                              P_("Fog"),
+                              P_("Settings for the depth cueing"),
                               CLUTTER_TYPE_FOG,
                               CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_FOG, pspec);
@@ -1114,9 +1249,8 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_boolean ("use-alpha",
-                                "Use Alpha",
-                                "Whether to honour the alpha component of "
-                                "the stage color",
+                                P_("Use Alpha"),
+                                P_("Whether to honour the alpha component of the stage color"),
                                 FALSE,
                                 CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_USE_ALPHA, pspec);
@@ -1132,12 +1266,29 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 1.2
    */
   pspec = g_param_spec_object ("key-focus",
-                               "Key Focus",
-                               "The currently key focused actor",
+                               P_("Key Focus"),
+                               P_("The currently key focused actor"),
                                CLUTTER_TYPE_ACTOR,
                                CLUTTER_PARAM_READWRITE);
   g_object_class_install_property (gobject_class, PROP_KEY_FOCUS, pspec);
 
+  /**
+   * ClutterStage:no-clear-hint:
+   *
+   * Whether or not the #ClutterStage should clear its contents
+   * before each paint cycle.
+   *
+   * See clutter_stage_set_no_clear_hint() for further information.
+   *
+   * Since: 1.4
+   */
+  pspec = g_param_spec_boolean ("no-clear-hint",
+                                P_("No Clear Hint"),
+                                P_("Whether the stage should clear its contents"),
+                                FALSE,
+                                CLUTTER_PARAM_READWRITE);
+  g_object_class_install_property (gobject_class, PROP_NO_CLEAR_HINT, pspec);
+
   /**
    * ClutterStage::fullscreen
    * @stage: the stage which was fullscreened
@@ -1147,7 +1298,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.6
    */
   stage_signals[FULLSCREEN] =
-    g_signal_new ("fullscreen",
+    g_signal_new (I_("fullscreen"),
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_FIRST,
 		  G_STRUCT_OFFSET (ClutterStageClass, fullscreen),
@@ -1164,7 +1315,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.6
    */
   stage_signals[UNFULLSCREEN] =
-    g_signal_new ("unfullscreen",
+    g_signal_new (I_("unfullscreen"),
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
 		  G_STRUCT_OFFSET (ClutterStageClass, unfullscreen),
@@ -1181,7 +1332,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.6
    */
   stage_signals[ACTIVATE] =
-    g_signal_new ("activate",
+    g_signal_new (I_("activate"),
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
 		  G_STRUCT_OFFSET (ClutterStageClass, activate),
@@ -1198,7 +1349,7 @@ clutter_stage_class_init (ClutterStageClass *klass)
    * Since: 0.6
    */
   stage_signals[DEACTIVATE] =
-    g_signal_new ("deactivate",
+    g_signal_new (I_("deactivate"),
 		  G_TYPE_FROM_CLASS (gobject_class),
 		  G_SIGNAL_RUN_LAST,
 		  G_STRUCT_OFFSET (ClutterStageClass, deactivate),
@@ -1256,9 +1407,10 @@ clutter_stage_init (ClutterStage *self)
 {
   ClutterStagePrivate *priv;
   ClutterBackend *backend;
+  ClutterGeometry geom;
 
   /* a stage is a top-level object */
-  CLUTTER_SET_PRIVATE_FLAGS (self, CLUTTER_ACTOR_IS_TOPLEVEL);
+  CLUTTER_SET_PRIVATE_FLAGS (self, CLUTTER_IS_TOPLEVEL);
 
   self->priv = priv = CLUTTER_STAGE_GET_PRIVATE (self);
 
@@ -1291,6 +1443,13 @@ clutter_stage_init (ClutterStage *self)
   priv->perspective.z_near = 0.1;
   priv->perspective.z_far  = 100.0;
 
+  cogl_matrix_init_identity (&priv->projection);
+  cogl_matrix_perspective (&priv->projection,
+                           priv->perspective.fovy,
+                           priv->perspective.aspect,
+                           priv->perspective.z_near,
+                           priv->perspective.z_far);
+
   /* depth cueing */
   priv->fog.z_near = 1.0;
   priv->fog.z_far  = 2.0;
@@ -1303,6 +1462,12 @@ clutter_stage_init (ClutterStage *self)
                     G_CALLBACK (clutter_stage_notify_min_size), NULL);
   g_signal_connect (self, "notify::min-height",
                     G_CALLBACK (clutter_stage_notify_min_size), NULL);
+
+  _clutter_stage_window_get_geometry (priv->impl, &geom);
+  _clutter_stage_set_viewport (self, 0, 0, geom.width, geom.height);
+
+  _clutter_stage_set_pick_buffer_valid (self, FALSE);
+  _clutter_stage_reset_picks_per_frame_counter (self);
 }
 
 /**
@@ -1406,18 +1571,30 @@ clutter_stage_set_perspective (ClutterStage       *stage,
 
   priv = stage->priv;
 
+  if (priv->perspective.fovy == perspective->fovy &&
+      priv->perspective.aspect == perspective->aspect &&
+      priv->perspective.z_near == perspective->z_near &&
+      priv->perspective.z_far == perspective->z_far)
+    return;
+
   priv->perspective = *perspective;
 
-  /* this will cause the viewport to be reset; see
-   * clutter_maybe_setup_viewport() inside clutter-main.c
-   */
-  CLUTTER_SET_PRIVATE_FLAGS (stage, CLUTTER_ACTOR_SYNC_MATRICES);
+  cogl_matrix_init_identity (&priv->projection);
+  cogl_matrix_perspective (&priv->projection,
+                           priv->perspective.fovy,
+                           priv->perspective.aspect,
+                           priv->perspective.z_near,
+                           priv->perspective.z_far);
+
+  priv->dirty_projection = TRUE;
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (stage));
 }
 
 /**
  * clutter_stage_get_perspective:
  * @stage: A #ClutterStage
- * @perspective: return location for a #ClutterPerspective
+ * @perspective: (out caller-allocates) (allow-none): return location for a
+ *   #ClutterPerspective
  *
  * Retrieves the stage perspective.
  */
@@ -1431,6 +1608,148 @@ clutter_stage_get_perspective (ClutterStage       *stage,
   *perspective = stage->priv->perspective;
 }
 
+/*
+ * clutter_stage_get_projection_matrix:
+ * @stage: A #ClutterStage
+ * @projection: return location for a #CoglMatrix representing the
+ *              perspective projection applied to actors on the given
+ *              @stage.
+ *
+ * Retrieves the @stage's projection matrix. This is derived from the
+ * current perspective set using clutter_stage_set_perspective().
+ *
+ * Since: 1.6
+ */
+void
+_clutter_stage_get_projection_matrix (ClutterStage *stage,
+                                      CoglMatrix *projection)
+{
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+  g_return_if_fail (projection != NULL);
+
+  *projection = stage->priv->projection;
+}
+
+/* This simply provides a simple mechanism for us to ensure that
+ * the projection matrix gets re-asserted before painting.
+ *
+ * This is used when switching between multiple stages */
+void
+_clutter_stage_dirty_projection (ClutterStage *stage)
+{
+  stage->priv->dirty_projection = TRUE;
+}
+
+/*
+ * clutter_stage_set_viewport:
+ * @stage: A #ClutterStage
+ * @x: The X postition to render the stage at, in window coordinates
+ * @y: The Y position to render the stage at, in window coordinates
+ * @width: The width to render the stage at, in window coordinates
+ * @height: The height to render the stage at, in window coordinates
+ *
+ * Sets the stage viewport. The viewport defines a final scale and
+ * translation of your rendered stage and actors. This lets you render
+ * your stage into a subregion of the stage window or you could use it to
+ * pan a subregion of the stage if your stage window is smaller then
+ * the stage. (XXX: currently this isn't possible)
+ *
+ * Unlike a scale and translation done using the modelview matrix this
+ * is done after everything has had perspective projection applied, so
+ * for example if you were to pan across a subregion of the stage using
+ * the viewport then you would not see a change in perspective for the
+ * actors on the stage.
+ *
+ * Normally the stage viewport will automatically track the size of the
+ * stage window with no offset so the stage will fill your window. This
+ * behaviour can be changed with the "viewport-mimics-window" property
+ * which will automatically be set to FALSE if you use this API. If
+ * you want to revert to the original behaviour then you should set
+ * this property back to %TRUE using
+ * clutter_stage_set_viewport_mimics_window().
+ * (XXX: If we were to make this API public then we might want to do
+ *  add that property.)
+ *
+ * Since: 1.6
+ */
+void
+_clutter_stage_set_viewport (ClutterStage *stage,
+                             int           x,
+                             int           y,
+                             int           width,
+                             int           height)
+{
+  ClutterStagePrivate *priv;
+
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  priv = stage->priv;
+
+
+  if (x == priv->viewport[0] &&
+      y == priv->viewport[1] &&
+      width == priv->viewport[2] &&
+      height == priv->viewport[3])
+    return;
+
+  priv->viewport[0] = x;
+  priv->viewport[1] = y;
+  priv->viewport[2] = width;
+  priv->viewport[3] = height;
+
+  priv->dirty_viewport = TRUE;
+
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (stage));
+}
+
+/* This simply provides a simple mechanism for us to ensure that
+ * the viewport gets re-asserted before next painting.
+ *
+ * This is used when switching between multiple stages */
+void
+_clutter_stage_dirty_viewport (ClutterStage *stage)
+{
+  stage->priv->dirty_viewport = TRUE;
+}
+
+/*
+ * clutter_stage_get_viewport:
+ * @stage: A #ClutterStage
+ * @x: A location for the X position where the stage is rendered,
+ *     in window coordinates.
+ * @y: A location for the Y position where the stage is rendered,
+ *     in window coordinates.
+ * @width: A location for the width the stage is rendered at,
+ *         in window coordinates.
+ * @height: A location for the height the stage is rendered at,
+ *          in window coordinates.
+ *
+ * Returns the viewport offset and size set using
+ * clutter_stage_set_viewport() or if the "viewport-mimics-window" property
+ * is TRUE then @x and @y will be set to 0 and @width and @height will equal
+ * the width if the stage window.
+ *
+ * Since: 1.6
+ */
+void
+_clutter_stage_get_viewport (ClutterStage *stage,
+                             int          *x,
+                             int          *y,
+                             int          *width,
+                             int          *height)
+{
+  ClutterStagePrivate *priv;
+
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  priv = stage->priv;
+
+  *x = priv->viewport[0];
+  *y = priv->viewport[1];
+  *width = priv->viewport[2];
+  *height = priv->viewport[3];
+}
+
 /**
  * clutter_stage_set_fullscreen:
  * @stage: a #ClutterStage
@@ -1476,6 +1795,15 @@ clutter_stage_set_fullscreen (ClutterStage *stage,
       if (iface->set_fullscreen)
 	iface->set_fullscreen (impl, fullscreen);
     }
+
+  /* If the backend did fullscreen the stage window then we need to resize
+   * the stage and update its viewport so we queue a relayout.  Note: if the
+   * fullscreen request is handled asynchronously we can't rely on this
+   * queue_relayout to update the viewport, but for example the X backend
+   * will recieve a ConfigureNotify after a successful resize which is how
+   * we ensure the viewport is updated on X.
+   */
+  clutter_actor_queue_relayout (CLUTTER_ACTOR (stage));
 }
 
 /**
@@ -2025,11 +2353,11 @@ clutter_stage_set_use_fog (ClutterStage *stage,
  *
  *     /&ast; set the fog color to the stage background color &ast;/
  *     clutter_stage_get_color (CLUTTER_STAGE (actor), &amp;stage_color);
- *     cogl_color_set_from_4ub (&amp;fog_color,
- *                              stage_color.red,
- *                              stage_color.green,
- *                              stage_color.blue,
- *                              stage_color.alpha);
+ *     cogl_color_init_from_4ub (&amp;fog_color,
+ *                               stage_color.red,
+ *                               stage_color.green,
+ *                               stage_color.blue,
+ *                               stage_color.alpha);
  *
  *     /&ast; enable fog &ast;/
  *     cogl_set_fog (&amp;fog_color,
@@ -2039,15 +2367,13 @@ clutter_stage_set_use_fog (ClutterStage *stage,
  *   }
  * ]|
  *
- * Note: The fogging functions only work correctly when the visible actors use
- *       unmultiplied alpha colors. By default Cogl will premultiply textures
- *       and cogl_set_source_color will premultiply colors, so unless you
- *       explicitly load your textures requesting an unmultiplied
- *       internal_format and use cogl_material_set_color you can only use
- *       fogging with fully opaque actors.
- *
- *       We can look to improve this in the future when we can depend on
- *       fragment shaders.
+ * <note>The fogging functions only work correctly when the visible actors use
+ * unmultiplied alpha colors. By default Cogl will premultiply textures and
+ * cogl_set_source_color() will premultiply colors, so unless you explicitly
+ * load your textures requesting an unmultiplied internal format and use
+ * cogl_material_set_color() you can only use fogging with fully opaque actors.
+ * Support for premultiplied colors will improve in the future when we can
+ * depend on fragment shaders.</note>
  *
  * Since: 0.6
  */
@@ -2223,11 +2549,33 @@ clutter_stage_ensure_viewport (ClutterStage *stage)
 {
   g_return_if_fail (CLUTTER_IS_STAGE (stage));
 
-  CLUTTER_SET_PRIVATE_FLAGS (stage, CLUTTER_ACTOR_SYNC_MATRICES);
+  _clutter_stage_dirty_viewport (stage);
 
   clutter_actor_queue_redraw (CLUTTER_ACTOR (stage));
 }
 
+void
+_clutter_stage_maybe_setup_viewport (ClutterStage *stage)
+{
+  ClutterStagePrivate *priv = stage->priv;
+
+  if (priv->dirty_viewport)
+    {
+      CLUTTER_NOTE (PAINT,
+                    "Setting up the viewport { w:%d, h:%d }",
+                    priv->viewport[2], priv->viewport[3]);
+
+      cogl_set_viewport (priv->viewport[0],
+                         priv->viewport[1],
+                         priv->viewport[2],
+                         priv->viewport[3]);
+
+    }
+
+  if (priv->dirty_projection)
+    cogl_set_projection_matrix (&priv->projection);
+}
+
 /**
  * clutter_stage_ensure_redraw:
  * @stage: a #ClutterStage
@@ -2522,7 +2870,7 @@ _clutter_stage_get_pending_swaps (ClutterStage *stage)
 {
   ClutterStageWindow *stage_window;
 
-  if (CLUTTER_PRIVATE_FLAGS (stage) & CLUTTER_ACTOR_IN_DESTRUCTION)
+  if (CLUTTER_ACTOR_IN_DESTRUCTION (stage))
     return 0;
 
   stage_window = _clutter_stage_get_window (stage);
@@ -2531,3 +2879,110 @@ _clutter_stage_get_pending_swaps (ClutterStage *stage)
 
   return _clutter_stage_window_get_pending_swaps (stage_window);
 }
+
+/**
+ * clutter_stage_set_no_clear_hint:
+ * @stage: a #ClutterStage
+ * @no_clear: %TRUE if the @stage should not clear itself on every
+ *   repaint cycle
+ *
+ * Sets whether the @stage should clear itself at the beginning
+ * of each paint cycle or not.
+ *
+ * Clearing the #ClutterStage can be a costly operation, especially
+ * if the stage is always covered - for instance, in a full-screen
+ * video player or in a game with a background texture.
+ *
+ * <note><para>This setting is a hint; Clutter might discard this
+ * hint depending on its internal state.</para></note>
+ *
+ * <warning><para>If parts of the stage are visible and you disable
+ * clearing you might end up with visual artifacts while painting the
+ * contents of the stage.</para></warning>
+ *
+ * Since: 1.4
+ */
+void
+clutter_stage_set_no_clear_hint (ClutterStage *stage,
+                                 gboolean      no_clear)
+{
+  ClutterStagePrivate *priv;
+  ClutterStageHint new_hints;
+
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  priv = stage->priv;
+  new_hints = priv->stage_hints;
+
+  if (no_clear)
+    new_hints |= CLUTTER_STAGE_NO_CLEAR_ON_PAINT;
+  else
+    new_hints &= ~CLUTTER_STAGE_NO_CLEAR_ON_PAINT;
+
+  if (priv->stage_hints == new_hints)
+    return;
+
+  priv->stage_hints = new_hints;
+
+  g_object_notify (G_OBJECT (stage), "no-clear-hint");
+}
+
+/**
+ * clutter_stage_get_no_clear_hint:
+ * @stage: a #ClutterStage
+ *
+ * Retrieves the hint set with clutter_stage_set_no_clear_hint()
+ *
+ * Return value: %TRUE if the stage should not clear itself on every paint
+ *   cycle, and %FALSE otherwise
+ *
+ * Since: 1.4
+ */
+gboolean
+clutter_stage_get_no_clear_hint (ClutterStage *stage)
+{
+  g_return_val_if_fail (CLUTTER_IS_STAGE (stage), FALSE);
+
+  return (stage->priv->stage_hints & CLUTTER_STAGE_NO_CLEAR_ON_PAINT) != 0;
+}
+
+gboolean
+_clutter_stage_get_pick_buffer_valid (ClutterStage *stage)
+{
+  g_return_val_if_fail (CLUTTER_IS_STAGE (stage), FALSE);
+
+  return stage->priv->have_valid_pick_buffer;
+}
+
+void
+_clutter_stage_set_pick_buffer_valid (ClutterStage *stage,
+                                      gboolean      valid)
+{
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  stage->priv->have_valid_pick_buffer = !!valid;
+}
+
+void
+_clutter_stage_increment_picks_per_frame_counter (ClutterStage *stage)
+{
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  stage->priv->picks_per_frame++;
+}
+
+void
+_clutter_stage_reset_picks_per_frame_counter (ClutterStage *stage)
+{
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  stage->priv->picks_per_frame = 0;
+}
+
+guint
+_clutter_stage_get_picks_per_frame_counter (ClutterStage *stage)
+{
+  g_return_val_if_fail (CLUTTER_IS_STAGE (stage), 0);
+
+  return stage->priv->picks_per_frame;
+}
