@@ -15,9 +15,6 @@ GType clutter_x11_filter_return_get_type (void) G_GNUC_CONST;
 GType clutter_x11_xinput_event_types_get_type (void) G_GNUC_CONST;
 #define CLUTTER_X11_TYPE_X11_XINPUT_EVENT_TYPES (clutter_x11_xinput_event_types_get_type())
 
-GType clutter_x11_input_device_type_get_type (void) G_GNUC_CONST;
-#define CLUTTER_X11_TYPE_X11_INPUT_DEVICE_TYPE (clutter_x11_input_device_type_get_type())
-
 G_END_DECLS
 
 #endif /* !__CLUTTER_X11_ENUM_TYPES_H__ */
